GtkLabel.tray_label,
GtkLabel.current_label {
  font-weight: bold;
}

GtkLabel.fullscreen_label {
  font-size: 19.5px;/*= 20000 / 1024 */
}

GtkLabel.fullscreen_label2 {
  font-size: 11.7px;/*= 12000/1024 */
}

GtkLabel.fullscreen_label,
GtkLabel.fullscreen_label2 {
  color: white;
}

